/* (c) Shereef Marzouk. See "licence DDRace.txt" and the readme.txt in the root of the distribution for more information. */
/* This is used for all NONE Official Builds of DDRace, if you changed the source in anyway change this to something other than DDRace */
#ifndef GAME_SERVER_GAMEMODES_GAMEMODE_H
#define GAME_SERVER_GAMEMODES_GAMEMODE_H

#define GAME_NAME "DDraceNetwork"
#define TEST_NAME "TestDDraceNetwork"
#endif // GAME_SERVER_GAMEMODES_GAMEMODE_H
